﻿using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.MetaOptimization {
  [Item("AlgorithmEvaluator", "")]
  [StorableClass]
  public class AlgorithmEvaluator : SingleSuccessorOperator {

    #region Parameter properties
    public ILookupParameter<IRandom> RandomParameter {
      get { return (LookupParameter<IRandom>)Parameters["Random"]; }
    }
    public ILookupParameter<IAlgorithm> AlgorithmParameter {
      get { return (LookupParameter<IAlgorithm>)Parameters["Algorithm"]; }
    }
    public ILookupParameter<IntValue> ProblemIndexParameter {
      get { return (LookupParameter<IntValue>)Parameters["ProblemIndex"]; }
    }
    public ILookupParameter<IntValue> RepetitionIndexParameter {
      get { return (LookupParameter<IntValue>)Parameters["RepetitionIndex"]; }
    }
    public ILookupParameter<ParameterConfigurationTree> ParameterConfigurationParameter {
      get { return (ILookupParameter<ParameterConfigurationTree>)Parameters["ParameterConfigurationTree"]; }
    }
    //public IValueParameter<IItemList<IRun>> RunsParameter {
    //  get { return (IValueParameter<IItemList<IRun>>)Parameters["Runs"]; }
    //}
    public LookupParameter<ResultCollection> ResultsParameter {
      get { return (LookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    #endregion

    [StorableConstructor]
    protected AlgorithmEvaluator(bool deserializing) : base(deserializing) { }
    public AlgorithmEvaluator()
      : base() {
      Parameters.Add(new LookupParameter<IRandom>("Random", "The pseudo random number generator which should be used to initialize the new random permutation."));
      Parameters.Add(new LookupParameter<IAlgorithm>("Algorithm", ""));
      Parameters.Add(new LookupParameter<IntValue>("ProblemIndex", ""));
      Parameters.Add(new LookupParameter<IntValue>("RepetitionIndex", ""));
      Parameters.Add(new LookupParameter<ParameterConfigurationTree>("ParameterConfigurationTree", ""));
      Parameters.Add(new LookupParameter<ResultCollection>("Results", ""));
    }
    protected AlgorithmEvaluator(AlgorithmEvaluator original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new AlgorithmEvaluator(this, cloner);
    }

    public override IOperation Apply() {
      bool isValidScope = RepetitionIndexParameter.ActualValue != null; // this is a workaround for OSGA, where the PMOEvaluator is executed on scopes which happen to contain additional subscopes which collide with the subscopes created for the repetitions/problems in PMOEvaluator
      if (!isValidScope) return base.Apply();

      ItemDictionary<StringValue, RunCollection> solutionCache = ResultsParameter.ActualValue.ContainsKey("SolutionCache") ? (ItemDictionary<StringValue, RunCollection>)ResultsParameter.ActualValue["SolutionCache"].Value : null;
      ParameterConfigurationTree parameterConfiguration = ParameterConfigurationParameter.ActualValue;
      IAlgorithm algorithm = AlgorithmParameter.ActualValue;
      int repetitionIndex = RepetitionIndexParameter.ActualValue.Value;

      if (solutionCache != null && solutionCache.Count(x => x.Key.Value == parameterConfiguration.ParameterInfoString) > 0) {
        algorithm.Runs.Add(solutionCache.Single(x => x.Key.Value == parameterConfiguration.ParameterInfoString).Value.ElementAt(repetitionIndex));
      } else {
        algorithm.StartSync(CancellationToken);
      }
      return base.Apply();
    }

    // for debug purposes, remove later
    private static string PrintGrammar(ISymbolicExpressionGrammar grammar) {
      return string.Join(string.Empty, grammar.Symbols.Select(x => x.InitialFrequency).ToArray());
      //foreach (var symbol in grammar.Symbols) {
      //  Console.WriteLine("{0} ({1})", symbol.ToString(), symbol.InitialFrequency);
      //}
    }
  }
}
